import unicodedata

from .vocabulary import Vocabulary
from math import log
import numpy as np
from .util import tr_lower, load_stopwords, deprecate, ConfigNotSet, VocabularyIsNotSet, WordVectorNotFound

IDF_SMOOTH, IDF_PROBABILISTIC = "smooth", "probabilistic"
IDF_METHOD_VALUES = [IDF_SMOOTH, IDF_PROBABILISTIC]


class IDFImpl:
    def __init__(self):
        pass

    def get_idf(self, method=IDF_SMOOTH, drop_stopwords=False, lowercase=False, drop_suffix=False, drop_punct=False,
                **kwargs):

        if method not in IDF_METHOD_VALUES:
            raise ValueError(f"Unknown idf method ({method}). Choose one of {IDF_METHOD_VALUES}")

        if lowercase:
            v = np.zeros(self.vocabulary.size)
        else:
            v = np.zeros(self.vocabulary.size_cs)

        if lowercase:
            tokens = [tr_lower(t) for t in self.tokens]
        else:
            tokens = self.tokens

        for token in tokens:
            t = Token(token)
            if t.is_oov or (drop_stopwords and t.is_stopword) or (drop_suffix and t.is_suffix) or (
                    drop_punct and t.is_punct):
                continue

            if lowercase:
                if method == IDF_SMOOTH:
                    v[t.id] = t.smooth_idf
                else:
                    v[t.id] = t.prob_idf
            else:
                if method == IDF_SMOOTH:
                    v[t.id_cs] = t.smooth_idf_cs
                else:
                    v[t.id_cs] = t.prob_idf_cs

        return v

    @property
    def idf(self):
        idf = self.config['idf']['method']
        drop_stopwords = self.config['default'].getboolean('drop_stopwords')
        lowercase = self.config['default'].getboolean('lowercase')
        drop_suffix = self.config['bert'].getboolean('drop_suffix')
        drop_punct = self.config['default'].getboolean('drop_punct')

        return self.get_idf(idf, drop_stopwords, lowercase, drop_suffix, drop_punct)


def word_shape(text):
    if len(text) >= 100:
        return "LONG"
    shape = []
    last = ""
    shape_char = ""
    seq = 0
    for char in text:
        if char.isalpha():
            if char.isupper():
                shape_char = "X"
            else:
                shape_char = "x"
        elif char.isdigit():
            shape_char = "d"
        else:
            shape_char = char
        if shape_char == last:
            seq += 1
        else:
            seq = 0
            last = shape_char
        if seq < 4:
            shape.append(shape_char)
    return "".join(shape)


class Token:
    config = None
    STOPWORDS = set(load_stopwords())
    vocabulary = None
    cache = {}

    @classmethod
    def _create_token(cls, word: str):
        token = object.__new__(cls)

        token.word = word
        token.lower_ = tr_lower(token.word)
        token.is_punct = all(unicodedata.category(c).startswith("P") for c in token.word)
        token.is_digit = token.word.isdigit()
        token.is_suffix = token.word.startswith('##')
        token.shape = word_shape(token.word)

        return token

    def __new__(cls, word: str):

        if word not in cls.cache:
            cls.cache[word] = cls._create_token(word)

        return cls.cache[word]

    @classmethod
    def set_vocabulary(cls, vocab: Vocabulary):
        Token.vocabulary = vocab
        Token.cache.clear()

    @classmethod
    def set_config(cls, config):
        Token.config = config

    @property
    def entry(self):
        deprecate("entry property is deprecated", (0, 20))

        return self

    @property
    def idf(self):
        if Token.config is None:
            raise ConfigNotSet("First run set_config.")
        else:
            if Token.config['idf']['method'] == IDF_SMOOTH:
                return self.smooth_idf
            else:
                return self.prob_idf

    @property
    def smooth_idf(self):
        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            return log(self.vocabulary.document_count / (1 + self.df)) + 1

    @property
    def smooth_idf_cs(self):
        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            return log(self.vocabulary.document_count / (1 + self.df_cs)) + 1

    @property
    def prob_idf(self) -> float:
        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            df = self.df + 1
            return log((self.vocabulary.document_count - df) / df)

    @property
    def prob_idf_cs(self) -> float:
        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            df = self.df_cs + 1
            return log((self.vocabulary.document_count - df) / df)

    @property
    def id_cs(self) -> int:
        """Vocabulary identifier"""

        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            return self.vocabulary.id_cs(self.word, -1)

    @property
    def id(self) -> int:
        """Vocabulary identifier (incase sensitive, aka lowercase all tokens)"""

        if Token.vocabulary is None:
            raise VocabularyIsNotSet("First run set_vocabulary")
        else:
            return self.vocabulary.id(self.word, -1)

    @property
    def is_oov(self) -> bool:
        """Return False if token type is not a part of vocabulary"""

        return self.id == -1

    @property
    def is_stopword(self) -> bool:
        return self.lower_ in Token.STOPWORDS

    @property
    def df(self) -> int:
        return self.vocabulary.df(self.word)

    @property
    def df_cs(self) -> int:
        """case sensitive document frequency"""
        return self.vocabulary.df_cs(self.word)

    @property
    def has_vector(self) -> bool:
        return self.vocabulary.has_vector(self.word)

    @property
    def vector(self) -> np.ndarray:
        if self.has_vector:
            return self.vocabulary.vector(self.word)
        else:
            raise WordVectorNotFound(self.word)

    def __str__(self):
        return self.word

    def __repr__(self):
        return self.word



import re
import sys
from collections import Counter
from functools import partial
from typing import List

import numpy as np  # type:ignore
from rich.console import Console
from scipy.sparse import csr_matrix
from cached_property import cached_property

from .token import Token, IDF_METHOD_VALUES, IDFImpl
from .util import tr_lower, select_layer, __tr_lower_abbrv__, flatten, pad, normalize_tokenizer_name
from .word_tokenizer import WordTokenizer
from ..config import load_config
from ..metrics import rouge1_score
from ..ml.sbd import load_model

console = Console()


class Span:
    def __init__(self, i: int, span, doc):
        self.doc = doc
        self.i = i
        self.value = span

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text

    @property
    def text(self):
        return self.doc.raw[slice(*self.value)]

    def span_features(self):
        is_first_span = self.i == 0
        is_last_span = self.i == len(self.doc.spans) - 1
        word = self.doc.raw[slice(*self.value)]

        if not is_first_span:
            word_m1 = self.doc.raw[slice(*self.doc.spans[self.i - 1].value)]
        else:
            word_m1 = '<D>'

        if not is_last_span:
            word_p1 = self.doc.raw[slice(*self.doc.spans[self.i + 1].value)]
        else:
            word_p1 = '</D>'

        features = {}

        # All upper features
        if word_m1.isupper() and not is_first_span:
            features["PREV_ALL_CAPS"] = True

        if word.isupper():
            features["ALL_CAPS"] = True

        if word_p1.isupper() and not is_last_span:
            features["NEXT_ALL_CAPS"] = True

        # All lower features
        if word_m1.islower() and not is_first_span:
            features["PREV_ALL_LOWER"] = True

        if word.islower():
            features["ALL_LOWER"] = True

        if word_p1.islower() and not is_last_span:
            features["NEXT_ALL_LOWER"] = True

        # Century
        if tr_lower(word_p1).startswith("yüzyıl") and not is_last_span:
            features["NEXT_WORD_CENTURY"] = True

        # Number
        if (tr_lower(word_p1).startswith("milyar") or tr_lower(word_p1).startswith("milyon") or tr_lower(
                word_p1).startswith("bin")) and not is_last_span:
            features["NEXT_WORD_NUMBER"] = True

        # Percentage
        if tr_lower(word_m1).startswith("yüzde") and not is_first_span:
            features["PREV_WORD_PERCENTAGE"] = True

        # In parenthesis feature
        if word[0] == '(' and word[-1] == ')':
            features["IN_PARENTHESIS"] = True

        # Suffix features
        m = re.search(r'\W+$', word)

        if m:
            subspan = m.span()
            suff2 = word[slice(*subspan)][:2]

            if all((c in '!\').?’”":;…') for c in suff2):
                features['NON-ALNUM SUFFIX2'] = suff2

        if word_p1[0] in '-*◊' and not is_last_span:
            features["NEXT_BULLETIN"] = True

        # title features
        # if word_m1.istitle() and not is_first_span:
        #     features["PREV_TITLE"] = 1

        # NOTE: 'Beşiktaş’ın' is not title by python :/
        m = re.search(r'\w+', word)

        if m:
            subspan = m.span()
            if word[slice(*subspan)].istitle():
                features["TITLE"] = True

        m = re.search(r'\w+', word_p1)

        if m:
            subspan = m.span()

            if word_p1[slice(*subspan)].istitle() and not is_last_span:
                features["NEXT_TITLE"] = True

        # suffix features
        for name, symbol in zip(['DOT', 'ELLIPSES', 'ELLIPSES', 'EXCLAMATION', 'QUESTION'],
                                ['.', '...', '…', '?', '!']):
            if word.endswith(symbol):
                features[f"ENDS_WITH_{name}"] = True

        # prefix abbreviation features
        if '.' in word:

            if any(tr_lower(word).startswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features["PREFIX_IS_ABBRV"] = True
            if any(tr_lower(word).endswith(abbrv) for abbrv in __tr_lower_abbrv__):
                features["SUFFIX_IS_ABBRV"] = True
            else:
                prefix = word.split('.', maxsplit=1)[0]
                features["PREFIX_IS_DIGIT"] = prefix.isdigit()

        # if '.' in word_m1:
        #     prefix_m1 = word_m1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_m1) in __tr_lower_abbrv__ and not is_first_span:
        #         features[f"PREV_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["PREV_PREFIX_IS_DIGIT"] = prefix_m1.isdigit()
        #
        # if '.' in word_p1:
        #     prefix_p1 = word_p1.split('.', maxsplit=1)[0]
        #
        #     if tr_lower(prefix_p1) in __tr_lower_abbrv__ and not is_last_span:
        #         features[f"NEXT_PREFIX_IS_ABBRV"] = True
        #     else:
        #         features["NEXT_PREFIX_IS_DIGIT"] = prefix_p1.isdigit()

        return features


TF_BINARY, TF_RAW, TF_FREQ, TF_LOG_NORM, TF_DOUBLE_NORM = "binary", "raw", "freq", "log_norm", "double_norm"
TF_METHOD_VALUES = [TF_BINARY, TF_RAW, TF_FREQ, TF_LOG_NORM, TF_DOUBLE_NORM]


class TFImpl:
    def __init__(self):
        pass

    def raw_tf(self, drop_stopwords=False, lowercase=False, drop_suffix=False, drop_punct=False) -> np.ndarray:
        if lowercase:
            v = np.zeros(self.vocabulary.size)
        else:
            v = np.zeros(self.vocabulary.size_cs)

        if lowercase:
            tokens = [tr_lower(t) for t in self.tokens]
        else:
            tokens = self.tokens

        counter = Counter(tokens)

        for token in tokens:
            t = Token(token)
            if t.is_oov or (drop_stopwords and t.is_stopword) or (drop_suffix and t.is_suffix) or (
                    drop_punct and t.is_punct):
                continue

            if lowercase:
                v[t.id] = counter[token]
            else:
                v[t.id_cs] = counter[token]

        return v

    def binary_tf(self, drop_stopwords=False, lowercase=False, drop_prefix=False, drop_punct=False) -> np.ndarray:
        return self.raw_tf(drop_stopwords, lowercase, drop_prefix, drop_punct).clip(max=1)

    def freq_tf(self, drop_stopwords=False, lowercase=False, drop_prefix=False, drop_punct=False) -> np.ndarray:
        tf = self.raw_tf(drop_stopwords, lowercase, drop_prefix, drop_punct)

        normalization = tf.sum()

        if normalization > 0:
            return tf / normalization
        else:
            return tf

    def log_norm_tf(self, drop_stopwords=False, lowercase=False, drop_prefix=False, drop_punct=False) -> np.ndarray:
        return np.log1p(self.raw_tf(drop_stopwords, lowercase, drop_prefix, drop_punct))

    def double_norm_tf(self, drop_stopwords=False, lowercase=False, drop_prefix=False, drop_punct=False,
                       k=0.5) -> np.ndarray:
        if not (0 < k < 1):
            raise ValueError(f"Ensure that 0 < k < 1 for double normalization term frequency calculation ({k} given)")

        tf = self.raw_tf(drop_stopwords, lowercase, drop_prefix, drop_punct)
        normalization = tf.max()

        if normalization > 0:
            return k + (1 - k) * (tf / normalization)
        else:
            return tf

    def get_tf(self, method=TF_BINARY, drop_stopwords=False, lowercase=False, drop_suffix=False, drop_punct=False,
               **kwargs) -> np.ndarray:
        if method == TF_BINARY:
            return self.binary_tf(drop_stopwords, lowercase, drop_suffix, drop_punct)
        elif method == TF_RAW:
            return self.raw_tf(drop_stopwords, lowercase, drop_suffix, drop_punct)
        elif method == TF_FREQ:
            return self.freq_tf(drop_stopwords, lowercase, drop_suffix, drop_punct)
        elif method == TF_LOG_NORM:
            return self.log_norm_tf(drop_stopwords, lowercase, drop_suffix, drop_punct)
        elif method == TF_DOUBLE_NORM:
            return self.double_norm_tf(drop_stopwords, lowercase, drop_suffix, drop_punct, **kwargs)
        else:
            raise ValueError(f"Unknown tf method ({method}). Choose one of {TF_METHOD_VALUES}")

    @property
    def tf(self):
        tf = self.config['tf']['method']
        drop_stopwords = self.config['default'].getboolean('drop_stopwords')
        lowercase = self.config['default'].getboolean('lowercase')
        drop_suffix = self.config['bert'].getboolean('drop_suffix')
        drop_punct = self.config['default'].getboolean('drop_punct')

        return self.get_tf(tf, drop_stopwords, lowercase, drop_suffix, drop_punct)


class BM25Impl:
    def __init__(self):
        pass

    def get_bm25(self, tf_method: str, idf_method: str, k1: float, b: float, delta: float = 0, drop_stopwords=False,
                 lowercase=False,
                 drop_suffix=False,
                 drop_punct=False, **kwargs):
        """Return bm25 embedding

        Refer to https://en.wikipedia.org/wiki/Okapi_BM25 for details

        :param tf_method:
        :param idf_method:
        :param k1:
        :param b:
        :param delta:
        :param drop_stopwords:
        :param lowercase:
        :param drop_suffix:
        :param drop_punct:
        :param kwargs:
        :return:
        """

        tf = self.get_tf(tf_method, drop_stopwords, lowercase, drop_suffix, drop_punct, **kwargs)
        idf = self.get_idf(idf_method, drop_stopwords, lowercase, drop_suffix, drop_punct, **kwargs)

        bm25 = idf * ((tf * (k1 + 1)) / (tf + k1 * (1 - b + b * (len(self) / self.avgdl))) + delta)

        return bm25

    def get_bm11(self, tf_method, idf_method, k1, avgdl, delta=0, drop_stopwords=False, lowercase=False,
                 drop_suffix=False,
                 drop_punct=False, **kwargs):
        return self.get_bm25(tf_method, idf_method, k1, 1, avgdl, delta, drop_stopwords, lowercase, drop_suffix,
                             drop_punct,
                             kwargs)

    def get_bm15(self, tf_method, idf_method, k1, avgdl, delta=0, drop_stopwords=False, lowercase=False,
                 drop_suffix=False,
                 drop_punct=False, **kwargs):
        return self.get_bm25(tf_method, idf_method, k1, 0, avgdl, delta, drop_stopwords, lowercase, drop_suffix,
                             drop_punct,
                             kwargs)


class Sentences(TFImpl, IDFImpl, BM25Impl):

    def __init__(self, id_: int, text: str, doc, config: dict = {}):
        TFImpl.__init__(self)
        IDFImpl.__init__(self)
        BM25Impl.__init__(self)

        self.id = id_
        self.text = text

        self._tokens = None
        self.document = doc
        self.config = doc.builder.config
        self._bert = None

        # No fail back to config read in here because this will slow down sentence instantiation extremely.
        self.tf_method = config['tf']['method']

        if self.tf_method == TF_BINARY:
            self.f_tf = self.binary_tf
        elif self.tf_method == TF_RAW:
            self.f_tf = self.raw_tf
        elif self.tf_method == TF_FREQ:
            self.f_tf = self.freq_tf
        elif self.tf_method == TF_LOG_NORM:
            self.f_tf = self.log_norm_tf
        elif self.tf_method == TF_DOUBLE_NORM:
            k = config.getfloat('tf', 'double_norm_k')

            if not 0 < k < 1:
                raise ValueError(
                    f"Invalid k value {k} for double norm term frequency. Values should be between 0 and 1.")
            self.f_tf = partial(self.double_norm_tf, k=k)
        else:
            raise ValueError(
                f"Unknown term frequency method {self.tf_method}. Choose on of {','.join(TF_METHOD_VALUES)}")

    @property
    def avgdl(self) -> int:
        """Average number of tokens per sentence"""
        return self.config['default'].getfloat('avg_sentence_length')

    @property
    def tokenizer(self):
        return self.document.tokenizer

    @property
    def vocabulary(self):
        return self.tokenizer.vocabulary

    @classmethod
    def set_tf_function(cls, tf_type):
        raise DeprecationWarning("Function is depreciated.")

    @property
    def bert(self):
        return self._bert

    @bert.setter
    def bert(self, bert):
        self._bert = bert

    @property
    def input_ids(self):
        return self.tokenizer.convert_tokens_to_ids(self.tokens_with_special_symbols)

    @property
    def tokens(self):
        if self._tokens is None:
            self._tokens = self.tokenizer(self.text)

        return self._tokens

    @property
    def tokens_with_special_symbols(self):
        return ['[CLS]'] + self.tokens + ['[SEP]']

    def rouge1(self, metric):
        return rouge1_score(
            flatten([[tr_lower(token) for token in sent.tokens] for sent in self.document if sent.id != self.id]),
            [tr_lower(t) for t in self.tokens], metric)

    @property
    def bm25(self) -> np.float32:

        tf = self.config['tf']['method']
        idf = self.config['idf']['method']
        drop_stopwords = self.config['default'].getboolean('drop_stopwords')
        lowercase = self.config['default'].getboolean('lowercase')
        drop_suffix = self.config['bert'].getboolean('drop_suffix')
        drop_punct = self.config['default'].getboolean('drop_punct')

        k1 = self.config['bm25'].getfloat('k1')
        b = self.config['bm25'].getfloat('b')

        delta = self.config['bm25'].getfloat('delta')

        return np.sum(self.get_bm25(tf, idf, k1, b, delta, drop_stopwords, lowercase, drop_suffix, drop_punct),
                      dtype=np.float32)

    @property
    def tfidf(self):
        tf = self.config['tf']['method']
        idf = self.config['idf']['method']
        drop_stopwords = self.config['default'].getboolean('drop_stopwords')
        lowercase = self.config['default'].getboolean('lowercase')
        drop_suffix = self.config['bert'].getboolean('drop_suffix')
        drop_punct = self.config['default'].getboolean('drop_punct')

        return self.get_tf(tf, drop_stopwords, lowercase, drop_suffix, drop_punct) * self.get_idf(idf, drop_stopwords,
                                                                                                  lowercase,
                                                                                                  drop_suffix,
                                                                                                  drop_punct)

    def get_tfidf(self, tf_method, idf_method, drop_stopwords=False, lowercase=False, drop_suffix=False,
                  drop_punct=False, **kwargs) -> np.ndarray:
        return self.get_tf(tf_method, drop_stopwords, lowercase, drop_suffix, drop_punct, **kwargs) * self.get_idf(
            idf_method, drop_stopwords, lowercase, drop_suffix, drop_punct, **kwargs)

    @property
    def tf(self):
        return self.f_tf()

    @property
    def idf(self):
        v = np.zeros(len(self.vocabulary))

        for token in self.tokens:
            t = self.vocabulary[token]
            if not t.is_oov:
                v[t.id] = t.idf

        return v

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text

    def __len__(self):
        return len(self.tokens)

    def __eq__(self, s: str):
        return self.text == s  # no need for type checking, will return false for non-strings

    def __getitem__(self, token_ix):
        return Token(self.tokens[token_ix])

    def __iter__(self):
        for t in self.tokens:
            yield Token(t)


class Document(TFImpl, IDFImpl, BM25Impl):
    def __init__(self, raw, builder):
        TFImpl.__init__(self)
        IDFImpl.__init__(self)
        BM25Impl.__init__(self)

        self.raw = raw
        self.spans = []
        self._sents = []
        self._tokens = None
        self.builder = builder
        self.config = self.builder.config

    @property
    def avgdl(self) -> int:
        """Average number of tokens per document"""
        return self.config['default'].getfloat('avg_document_length')

    @property
    def tokens(self):
        if self._tokens is None:
            self._tokens = []
            for s in self:
                for t in s.tokens:
                    self._tokens.append(t)

        return self._tokens

    @property
    def vocabulary(self):
        return self.tokenizer.vocabulary

    @property
    def tokenizer(self):
        return self.builder.tokenizer

    def __getitem__(self, key):
        return self._sents[key]

    def __iter__(self):
        return iter(self._sents)

    def __str__(self):
        return self.raw

    def __repr__(self):
        return self.raw

    def __len__(self):
        return len(self._sents)

    def max_length(self):
        """Maximum length of a sentence including special symbols."""
        return max(len(s.tokens_with_special_symbols) for s in self._sents)

    def padded_matrix(self, return_numpy=False, return_mask=True):
        """Returns a 0 padded numpy.array or torch.tensor
              One row for each sentence
              One column for each token (pad 0 if length of sentence is shorter than the max length)
        :param return_numpy: Whether to return numpy.array or torch.tensor
        :param return_mask: Whether to return padding mask
        :return:
        """
        max_len = self.max_length()

        if not return_numpy:
            try:
                import torch
            except ImportError:
                console.print(
                    ("Error in importing transformers module. "
                     "Ensure that you run 'pip install sadedegel[bert]' to use BERT features."))
                sys.exit(1)

            mat = torch.tensor([pad(s.input_ids, max_len) for s in self])

            if return_mask:
                return mat, (mat > 0).to(int)
            else:
                return mat
        else:
            mat = np.array([pad(s.input_ids, max_len) for s in self])

            if return_mask:
                return mat, (mat > 0).astype(int)
            else:
                return mat

    @cached_property
    def bert_embeddings(self):
        try:
            import torch
            from transformers import BertModel
        except ImportError:
            console.print(
                ("Error in importing transformers module. "
                 "Ensure that you run 'pip install sadedegel[bert]' to use BERT features."))
            sys.exit(1)

        inp, mask = self.padded_matrix()

        if DocBuilder.bert_model is None:
            DocBuilder.bert_model = BertModel.from_pretrained("dbmdz/bert-base-turkish-cased",
                                                              output_hidden_states=True)
            DocBuilder.bert_model.eval()

        with torch.no_grad():
            outputs = DocBuilder.bert_model(inp, mask)

        twelve_layers = outputs[2][1:]

        return select_layer(twelve_layers, [11], return_cls=False)

    def get_tfidf(self, tf_method, idf_method, **kwargs):
        return self.get_tf(tf_method, **kwargs) * self.get_idf(idf_method, **kwargs)

    @property
    def tfidf(self):
        return self.tf * self.idf

    @property
    def tfidf_matrix(self):
        indptr = [0]
        indices = []
        data = []

        for i in range(len(self)):
            sent_embedding = self[i].tfidf

            for idx in sent_embedding.nonzero()[0]:
                indices.append(idx)
                data.append(sent_embedding[idx])

            indptr.append(len(indices))

        lowercase = self.config['default'].getboolean('lowercase')

        if lowercase:
            dim = self.vocabulary.size
        else:
            dim = self.vocabulary.size_cs

        m = csr_matrix((data, indices, indptr), dtype=np.float32, shape=(len(self), dim))

        return m

    def from_sentences(self, sentences: List[str]):
        return self.builder.from_sentences(sentences)


class DocBuilder:
    bert_model = None

    def __init__(self, **kwargs):

        self.config = load_config(kwargs)

        self.sbd = load_model()

        tokenizer_str = normalize_tokenizer_name(self.config['default']['tokenizer'])

        self.tokenizer = WordTokenizer.factory(tokenizer_str)

        Token.set_vocabulary(self.tokenizer.vocabulary)

        self.config['default']['avg_sentence_length'] = self.config[tokenizer_str]['avg_sentence_length']
        self.config['default']['avg_document_length'] = self.config[tokenizer_str]['avg_document_length']

        idf_method = self.config['idf']['method']

        if idf_method in IDF_METHOD_VALUES:
            Token.config = self.config
        else:
            raise ValueError(f"Unknown term frequency method {idf_method}. Choose on of {IDF_METHOD_VALUES}")

    def __call__(self, raw):

        if raw is not None:
            raw_stripped = raw.strip()
            _spans = [match.span() for match in re.finditer(r"\S+", raw_stripped)]

            d = Document(raw_stripped, self)
            d.spans = [Span(i, span, d) for i, span in enumerate(_spans)]

            if len(d.spans) > 0:
                y_pred = self.sbd.predict((span.span_features() for span in d.spans))
            else:
                y_pred = []

            eos_list = [end for (start, end), y in zip(_spans, y_pred) if y == 1]

            if len(eos_list) > 0:
                for i, eos in enumerate(eos_list):
                    if i == 0:
                        d._sents.append(Sentences(i, d.raw[:eos].strip(), d, self.config))
                    else:
                        d._sents.append(Sentences(i, d.raw[eos_list[i - 1] + 1:eos].strip(), d, self.config))

                if eos_list[-1] != len(raw_stripped):
                    d._sents.append(Sentences(len(d._sents), d.raw[eos_list[-1] + 1:len(raw_stripped)], d,
                                              self.config))
            else:
                d._sents.append(Sentences(0, d.raw.strip(), d, self.config))

        else:
            raise Exception(f"{raw} document text can't be None")

        return d

    def from_sentences(self, sentences: List[str]):
        raw = "\n".join(sentences)

        d = Document(raw, self)
        for i, s in enumerate(sentences):
            d._sents.append(Sentences(i, s, d, self.config))

        return d



from .util import grid_search
from tabulate import tabulate

from ...dataset import load_annotated_corpus
from ...config import config_context
from .parameter_space import bm25_parameter_space, tfidf_parameter_space, lexrank_parameter_space, \
    textrank_parameter_space
from ..bm25 import BM25Summarizer
from ..tf_idf import TFIDFSummarizer
from ..baseline import BandSummarizer, RandomSummarizer, PositionSummarizer, LengthSummarizer
from ..rouge import Rouge1Summarizer
from ..rank import TextRank, LexRankSummarizer

import click

summarizer_list = [(BM25Summarizer, bm25_parameter_space, 50),
                   (TFIDFSummarizer, tfidf_parameter_space, 50),
                   (BandSummarizer, lambda x: [dict(k=i) for i in range(2, 6)], 5),
                   (RandomSummarizer, lambda x: [{"normalize": True}], 1),
                   (PositionSummarizer, lambda x: [dict(mode="first"), dict(mode="last")], 2),
                   (
                       Rouge1Summarizer, lambda x: [dict(metric="precision"), dict(metric="recall"), dict(metric="f1")],
                       3),
                   (LengthSummarizer, lambda x: [dict(mode="token"), dict(mode="char")], 2),
                   (LexRankSummarizer, lexrank_parameter_space, 10),
                   (TextRank, textrank_parameter_space, 10)]


@click.group(help="SadedeGel summarizer commandline")
def cli():
    pass


@cli.command()
@click.option("-f", "--table-format", default="github")
@click.option("-t", "--tag", default=["extractive"], multiple=True)
@click.option("-d", "--debug", default=False)
@click.option("--tokenizer", default="bert")
def evaluate(table_format, tag, debug, tokenizer):
    """Evaluate various extractive summarizer using random hyper parameter sampling"""

    anno = load_annotated_corpus(False)
    relevance = [[doc['relevance']] for doc in anno]

    with config_context(tokenizer=tokenizer) as DocBuilder:
        docs = [DocBuilder.from_sentences(doc['sentences']) for doc in anno]

    table = []

    for Summarizer, parameter_space, n_trial in filter(lambda x: any((t in x[0].tags) for t in tag), summarizer_list):
        if 'bert' in Summarizer.tags and tokenizer != 'bert':
            continue

        scores = grid_search(relevance, docs, Summarizer, parameter_space(n_trial))

        table += [[method[0], method[1], scores[0], scores[1], scores[2]] for
                  method, scores in scores.items()]

    print(
        tabulate(sorted(table, key=lambda x: x[2], reverse=True)[:5]
                 , headers=['Method', "Parameter", 'ndcg(k=0.1)', 'ndcg(k=0.5)', 'ndcg(k=0.8)'],
                 tablefmt="github",
                 floatfmt=".4f"))

    print(
        tabulate(sorted(table, key=lambda x: x[3], reverse=True)[:5]
                 , headers=['Method', "Parameter", 'ndcg(k=0.1)', 'ndcg(k=0.5)', 'ndcg(k=0.8)'],
                 tablefmt="github",
                 floatfmt=".4f"))

    print(
        tabulate(sorted(table, key=lambda x: x[4], reverse=True)[:5]
                 , headers=['Method', "Parameter", 'ndcg(k=0.1)', 'ndcg(k=0.5)', 'ndcg(k=0.8)'],
                 tablefmt="github",
                 floatfmt=".4f"))


if __name__ == '__main__':
    cli()



from typing import List
import numpy as np  # type: ignore
from ._base import ExtractiveSummarizer
from ..bblock import Sentences


class Rouge1Summarizer(ExtractiveSummarizer):
    """Assign a higher importance score based on ROUGE1 score of the sentences within the document.

    metric : {'f1', 'precision','recall'}, default='f1'
        Metric to be used for ROUGE1 computation.

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return L2 normalized score vector.
    """

    tags = ExtractiveSummarizer.tags + ['self-supervised', 'ml']

    def __init__(self, metric='f1', normalize=True):
        super().__init__(normalize)
        if metric not in ['f1', 'precision', 'recall']:
            raise ValueError(f"mode should be one of 'f1', 'precision','recall'")

        self.metric = metric

    def _predict(self, sentences: List[Sentences]):
        return np.array([sent.rouge1(self.metric) for sent in sentences])



import gzip
import os.path
import sys
from itertools import tee
from pathlib import Path
from shutil import copyfileobj

import boto3
import click
from loguru import logger
from rich.console import Console
from smart_open import open

from ._core import load_tweet_sentiment_train, CORPUS_SIZE

console = Console()

logger.disable("sadedegel")


@click.group(help="Twitter Sentiment Dataset Commandline")
def cli():
    pass


@cli.command()
@click.option("--access-key", help="Access Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_access_key', ''))
@click.option("--secret-key", help="Secret Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_secret_key', ''))
@click.option("--data_home", '-d', help="Data home directory", default="~/.sadedegel_data")
def download(access_key, secret_key, data_home):
    """Download twitter sentiment corpus from cloud with your key."""

    data_home = Path(os.path.expanduser(data_home)) / 'tweet_sentiment'
    data_home.mkdir(parents=True, exist_ok=True)
    console.print(f"Data directory for twitter sentiment data {data_home}")

    transport_params = {
        'session': boto3.Session(aws_access_key_id=access_key,
                                 aws_secret_access_key=secret_key),
        'resource_kwargs': {
            'endpoint_url': 'https://storage.googleapis.com',
        }
    }

    url = f"s3://sadedegel/dataset/tweet_sentiment_train.csv.gz"

    with open(url, 'rb', transport_params=transport_params) as fp, gzip.open(data_home / os.path.basename(url),
                                                                             "wb") as wp:
        copyfileobj(fp, wp)


@cli.command()
def validate():
    """Sanity check on corpus
    """
    with console.status("[bold yellow]Validating train"):
        train = load_tweet_sentiment_train()

        train_clone, train = tee(train, 2)

        n_train = sum(1 for _ in train_clone)
        categories = set([row['sentiment_class'] for row in train])

        if n_train == CORPUS_SIZE:
            console.log("Cardinality check [yellow]DONE[/yellow]")
        else:
            console.log(f"Cardinality check [red]FAILED[/red]")
            console.log(f"|Tweet sentiment (train)| : {n_train} ({CORPUS_SIZE} expected)")
            sys.exit(1)

        if categories == {0, 1}:
            console.log("Label check [yellow]DONE[/yellow]")
        else:
            console.log("Class check [red]FAILED[/red]")
            console.log(f"\tTweet sentiment classes : {categories} ({set(['POSITIVE', 'NEGATIVE'])} expected)")
            sys.exit(1)


if __name__ == "__main__":
    cli()



import csv
import gzip
from pathlib import Path
from rich.console import Console

CLASS_VALUES = ['POSITIVE', 'NEGATIVE']
CORPUS_SIZE = 11117

console = Console()

__general_download_message__ = """Ensure that you have properly downloaded tweet sentiment corpus using

            python -m sadedegel.dataset.tweet_sentiment download --access-key xxx --secret-key xxxx

        Unfortunately due to data licensing issues we could not share data publicly. 
        Get in touch with sadedegel team to obtain a download key.
        """


def check_directory_structure(path: str) -> bool:
    base_dir = Path(path).expanduser()

    tweet_sentiment_dir = base_dir / 'tweet_sentiment'

    if not base_dir.exists():
        console.log(f"Dataset base directory ([bold red]{base_dir}[/bold red]) does not exist")

    elif not tweet_sentiment_dir.exists():
        console.log(
            f"Tweet sentiment directory ([bold red]{tweet_sentiment_dir}[/bold red]) does not exist")

    else:
        return True

    console.log(__general_download_message__)

    return False


def load_tweet_sentiment_train(data_home="~/.sadedegel_data"):
    if not check_directory_structure(data_home):
        raise Exception("Tweet Sentiment Corpus validation error")

    train_csv = Path(data_home).expanduser() / "tweet_sentiment"
    train_csv = train_csv / "tweet_sentiment_train.csv.gz"

    with gzip.open(train_csv, "rt") as csvfile:
        rd = csv.DictReader(csvfile)

        for rec in rd:
            yield dict(id=rec['text_uuid'], tweet=rec['text'], sentiment_class=CLASS_VALUES.index(rec['sentiment']))



from math import ceil
from os.path import dirname
from pathlib import Path

import numpy as np
from joblib import dump, load as jl_load
from rich.console import Console
from sklearn.svm import SVC
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold
from sklearn.utils import shuffle

from ..dataset.tweet_sentiment import load_tweet_sentiment_train, CORPUS_SIZE, CLASS_VALUES
from ..extension.sklearn import TfidfVectorizer, Text2Doc
from sklearn.pipeline import Pipeline

from itertools import islice

console = Console()


def empty_model():
    return Pipeline(
        [('text2doc', Text2Doc("icu")),
         ('tfidf', TfidfVectorizer(tf_method='freq', idf_method='smooth', drop_punct=False, drop_stopwords=False,
                                   lowercase=True, show_progress=True)),
         ('svc', SVC(C=0.3392899481481453, kernel="linear", random_state=42))
         ]
    )


def cv(k=3, max_instances=-1):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    if max_instances > 0:
        raw = islice(load_tweet_sentiment_train(), max_instances)
    else:
        raw = load_tweet_sentiment_train()

    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    BATCH_SIZE = 1000

    kf = KFold(n_splits=k)
    console.log(f"Corpus Size: {CORPUS_SIZE}")

    scores = []

    for train_indx, test_index in kf.split(df):
        train = df.iloc[train_indx]
        test = df.iloc[test_index]

        n_split = ceil(len(train) / BATCH_SIZE)
        console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

        batches = np.array_split(train, n_split)

        pipeline = empty_model()

        pipeline.fit(train.tweet, train.sentiment_class)

        y_pred = pipeline.predict(test.tweet)

        scores.append(f1_score(test.sentiment_class, y_pred, average="macro"))

        console.log(scores)


def build(max_instances=-1, save=True):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    if max_instances > 0:
        raw = islice(load_tweet_sentiment_train(), max_instances)
    else:
        raw = load_tweet_sentiment_train()

    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    BATCH_SIZE = 1000

    console.log(f"Corpus Size: {CORPUS_SIZE}")

    n_split = ceil(len(df) / BATCH_SIZE)
    console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

    batches = np.array_split(df, n_split)

    pipeline = empty_model()

    pipeline.fit(df.tweet, df.sentiment_class)

    console.log("Model build [green]DONE[/green]")

    if save:
        model_dir = Path(dirname(__file__)) / 'model'

        model_dir.mkdir(parents=True, exist_ok=True)

        pipeline.steps[0][1].Doc = None

        dump(pipeline, (model_dir / 'tweet_sentiment.joblib').absolute(), compress=('gzip', 9))


def load(model_name="tweet_sentiment"):
    return jl_load(Path(dirname(__file__)) / 'model' / f"{model_name}.joblib")


if __name__ == '__main__':
    build()



from os.path import dirname
from pathlib import Path

from sklearn.ensemble import RandomForestClassifier  # type: ignore
from sklearn.feature_extraction import DictVectorizer  # type: ignore
from sklearn.pipeline import Pipeline  # type: ignore
from joblib import dump, load  # type: ignore

from loguru import logger


class SentenceBoundaryDetector:
    def __init__(self, name: str):
        if name:
            self.model_file = (Path(dirname(__file__)) / 'model' / name).absolute()
        else:
            self.model_file = None
        self.pipeline = None

    def predict(self, spans):
        if self.pipeline is None:
            logger.info(f"Loading sbd model from {self.model_file}")
            self.pipeline = load(self.model_file)

        return self.pipeline.predict(spans)

    def fit(self, features, y):
        self.pipeline.fit(features, y)

    @classmethod
    def from_pipeline(cls, pipeline):
        sbd = SentenceBoundaryDetector(None)
        sbd.pipeline = pipeline

        return sbd


def create_model():
    """Creates a new sbd model detector."""
    pipeline = Pipeline([('feat', DictVectorizer()), ('dt', RandomForestClassifier())])

    return SentenceBoundaryDetector.from_pipeline(pipeline)


def save_model(model: SentenceBoundaryDetector, name="sbd.pickle"):
    model_file = (Path(dirname(__file__)) / 'model' / name).absolute()

    dump(model.pipeline, model_file)


def load_model(name="sbd.pickle"):
    return SentenceBoundaryDetector(name)



from os.path import dirname
from pathlib import Path

from joblib import dump, load as jl_load
from rich.console import Console

from sklearn.linear_model import SGDClassifier
from sklearn.metrics import f1_score
from sklearn.utils import shuffle

from ..extension.sklearn import BM25Vectorizer, Text2Doc

from ..dataset.profanity import load_offenseval_train, load_offenseval_test, \
    load_offenseval_test_label, CORPUS_SIZE
from ..extension.sklearn import OnlinePipeline

console = Console()


def empty_model():
    return OnlinePipeline(
        [('text2doc', Text2Doc("icu")),
         ('bm25', BM25Vectorizer(tf_method='binary', idf_method='smooth',
                                 drop_stopwords=False,
                                 drop_punct=True, lowercase=True, k1=0.02978859422550008, b=4.825535232935995e-05,
                                 delta=1.2258845996051257,
                                 show_progress=True)),
         ('sgd', SGDClassifier(alpha=0.012998795143949122,
                               loss="modified_huber", average=False, penalty="l2",
                               random_state=42, fit_intercept=True))]
    )


def build(save=True):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    raw = load_offenseval_train()
    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    console.log(f"Corpus Size: {CORPUS_SIZE}")

    pipeline = empty_model()

    pipeline.fit(df.tweet, df.profanity_class)

    evaluate(pipeline)

    console.log("Model build [green]DONE[/green]")

    if save:
        model_dir = Path(dirname(__file__)) / 'model'

        model_dir.mkdir(parents=True, exist_ok=True)

        pipeline.steps[0][1].Doc = None

        dump(pipeline, (model_dir / 'tweet_profanity_classification.joblib').absolute(), compress=('gzip', 9))

        console.log("Model save [green]DONE[/green]")


def load():
    return jl_load(Path(dirname(__file__)) / 'model' / 'tweet_profanity_classification.joblib')


def evaluate(model=None):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    if model is None:
        model = load()

    test = pd.DataFrame.from_records(load_offenseval_test())
    testLabel = pd.DataFrame.from_records(load_offenseval_test_label())

    test = test.merge(testLabel, on='id')

    y_pred = model.predict(test.tweet)

    console.log(f"Model test accuracy (f1-macro): {f1_score(test.profanity_class, y_pred, average='macro')}")


if __name__ == "__main__":
    build(save=True)



import numpy as np
from scipy.stats.distributions import uniform
from sklearn.model_selection import ParameterSampler


def bm25_parameter_space(n_trials):
    rng = np.random.RandomState(42)

    return ParameterSampler(
        dict(tf_method=["binary", "raw", "freq", "log_norm", "double_norm"]
             , idf_method=["smooth", "probabilistic"],
             drop_stopwords=[True, False],
             drop_suffix=[True, False], drop_punct=[True, False],
             lowercase=[True, False],
             k1=uniform(1.2, 2.0),
             b=uniform(0.5, 0.8), delta=uniform(0, 2)), n_iter=n_trials, random_state=rng)


def tfidf_parameter_space(n_trials):
    rng = np.random.RandomState(42)

    return ParameterSampler(
        dict(tf_method=["binary", "raw", "freq", "log_norm", "double_norm"]
             , idf_method=["smooth", "probabilistic"],
             drop_stopwords=[True, False],
             drop_suffix=[True, False], drop_punct=[True, False],
             lowercase=[True, False]), n_iter=n_trials, random_state=rng)


def lexrank_parameter_space(n_trials):
    rng = np.random.RandomState(42)

    return ParameterSampler(
        dict(tf_method=["binary", "raw", "freq", "log_norm", "double_norm"]
             , idf_method=["smooth", "probabilistic"],
             drop_stopwords=[True, False],
             drop_suffix=[True, False], drop_punct=[True, False],
             lowercase=[True, False], threshold=uniform(0, 1)), n_iter=n_trials, random_state=rng)


def textrank_parameter_space(n_trials):
    rng = np.random.RandomState(42)

    return ParameterSampler(
        dict(alpha=uniform(0.0001, 1)), n_iter=n_trials, random_state=rng)



import sys
import warnings
from collections import defaultdict
from os.path import dirname
from pathlib import Path
from typing import List

import numpy as np
from rich.console import Console

from ..about import __version__

console = Console()

__tr_upper__ = "ABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZ"
__tr_lower__ = "abcçdefgğhıijklmnoöprsştuüvyz"

__tr_lower_abbrv__ = ['hz.', 'dr.', 'prof.', 'doç.', 'org.', 'sn.', 'st.', 'mah.', 'mh.', 'sok.', 'sk.', 'alb.', 'gen.',
                      'av.', 'ist.', 'ank.', 'izm.', 'm.ö.', 'k.k.t.c.']


def tr_lower(s: str) -> str:
    return s.replace("I", "ı").replace("İ", "i").lower()


def tr_upper(s: str) -> str:
    return s.replace("i", "İ").upper()


def space_pad(token):
    return " " + token + " "


def pad(l, padded_length):
    return l + [0 for _ in range(padded_length - len(l))]


def flatten(l2: List[List]):
    flat = []
    for l in l2:
        for e in l:
            flat.append(e)

    return flat


def is_eos(span, sentences: List[str]) -> int:
    start = 0
    eos = []
    for s in sentences:
        idx = span.doc.raw.find(s, start) + len(s) - 1
        eos.append(idx)

        start = idx

    b, e = span.value

    for idx in eos:
        if b <= idx <= e:
            return 1

    return 0


def select_layer(bert_out: tuple, layers: List[int], return_cls: bool) -> np.ndarray:
    """Selects and averages layers from BERT output.

    Parameters:
        bert_out: tuple
            Tuple containing output of 12 intermediate layers after feeding a document.

        layers: List[int]
            List that contains which layer to choose. max = 11, min = 0.

        return_cls: bool
            Whether to use CLS token embedding as sentence embedding instead of averaging token embeddings.

    Returns:
        numpy.ndarray (n_sentences, embedding_size) Embedding size if default to 768.

    """
    n_layers = len(layers)
    n_sentences = bert_out[0].shape[0]
    n_tokens = bert_out[0].shape[1]

    if not (min(layers) > -1 and max(layers) < 12):
        raise Exception(f"Value for layer should be in 0-11 range")

    if return_cls:
        cls_matrix = np.zeros((n_layers, n_sentences, 768))
        l_ix = 0
        for l, layer in enumerate(bert_out):
            if l not in layers:
                continue
            else:
                l_ix = l_ix + 1
            for s, sentence in enumerate(layer):
                cls_tensor = sentence[0].numpy()
                cls_matrix[l_ix - 1, s, :] = cls_tensor
        layer_mean_cls = np.mean(cls_matrix, axis=0)

        return layer_mean_cls

    else:
        token_matrix = np.zeros((n_layers, n_sentences, n_tokens - 2, 768))
        for l, layer in enumerate(bert_out):
            l_ix = 0
            if l not in layers:
                continue
            else:
                l_ix = l_ix + 1
            for s, sentence in enumerate(layer):
                for t, token in enumerate(sentence[1:-1]):  # Exclude [CLS] and [SEP] embeddings
                    token_tensor = sentence[t].numpy()
                    token_matrix[l_ix - 1, s, t, :] = token_tensor

        tokenwise_mean = np.mean(token_matrix, axis=2)
        layer_mean_token = np.mean(tokenwise_mean, axis=0)

        return layer_mean_token


def normalize_tokenizer_name(tokenizer_name, raise_on_error=False):
    normalized = tokenizer_name.lower().replace(' ', '').replace('-', '').replace('tokenizer', '')

    if normalized not in ['bert', 'simple', 'icu']:
        msg = f"Invalid tokenizer {tokenizer_name} ({normalized}). Valid values are bert, simple"
        if raise_on_error:
            raise ValueError(msg)
        else:
            warnings.warn(msg, UserWarning, stacklevel=3)

    return normalized


def to_config_dict(kw: dict):
    d = defaultdict(lambda: dict())
    for k, v in kw.items():
        if '__' not in k:  # default section
            d['default'][k] = v
        else:
            section, key = k.split('__')

            d[section][key] = v

    return d


def load_stopwords(base_path=None):
    """ Return Turkish stopwords as list from file. """
    if base_path is None:
        base_path = dirname(__file__)

    text_path = Path(base_path) / "data" / "stop-words.txt"

    with open(text_path, "r") as fp:
        stopwords = fp.readlines()

    stopwords = [s.rstrip() for s in stopwords]

    return stopwords


def deprecate(message: str, eol_version: tuple):
    current = tuple([int(v) for v in __version__.split('.')])

    if current >= eol_version:
        console.print(f"[red]{message}[/red]")
        sys.exit(1)
    else:
        console.print(f"[magenta]{message}[/magenta], will be dropped by {'.'.join(map(str, eol_version))}")


class ConfigNotSet(Exception):
    """Used when configuration is not set although it is expected"""


class VocabularyIsNotSet(Exception):
    """Raised when vocabulary class variable is not set"""


class WordVectorNotFound(Exception):
    """Word Vector is not available for word"""

    def __init__(self, w: str):
        self.word = w

        super().__init__(f"Word Vector not found for {self.word}")



__all__ = ['tokenizer_context', 'config_context', 'idf_context', 'tf_context', 'load_config']

import warnings
from collections import defaultdict
from configparser import ConfigParser
from contextlib import contextmanager
from os.path import dirname
from pathlib import Path

from rich.console import Console
from rich.table import Table


@contextmanager
def tokenizer_context(tokenizer_name, warning=False):
    from .bblock import DocBuilder  # pylint: disable=import-outside-toplevel

    if warning:
        warnings.warn(f"Changing tokenizer to {tokenizer_name}")

    yield DocBuilder(tokenizer=tokenizer_name)


@contextmanager
def config_context(**kwargs):
    from .bblock import DocBuilder  # pylint: disable=import-outside-toplevel

    yield DocBuilder(**kwargs)


@contextmanager
def idf_context(idf_type, warning=False):  # pylint: disable=unused-argument
    from .bblock import DocBuilder  # pylint: disable=import-outside-toplevel

    yield DocBuilder(idf__method=idf_type)


@contextmanager
def tf_context(tf_type, warning=False):  # pylint: disable=unused-argument
    from .bblock import DocBuilder  # pylint: disable=import-outside-toplevel

    yield DocBuilder(tf__method=tf_type)


def to_config_dict(kw: dict):
    d = defaultdict(dict)
    for k, v in kw.items():
        if '__' not in k:  # default section
            d['default'][k] = v
        else:
            section, key = k.split('__')

            d[section][key] = v

    return d


def load_config(kwargs: dict = None):
    config = ConfigParser()
    config.read([Path(dirname(__file__)) / 'default.ini', Path("~/.sadedegel/user.ini").expanduser()])

    if kwargs:
        config_dict = to_config_dict(kwargs)
        config.read_dict(config_dict)

    return config


def show_config(config, section=None):
    descriptions = {"default__tokenizer": "Word tokenizer to use",
                    "default__drop_stopwords": ("Whether to drop stopwords in various calculations. "
                                                "Such as, tfidf, bm25, etc."),
                    "default__lowercase": "Whether to use lowercased form rather than form itself.",
                    "default__drop_punct": ("Whether to drop punctuations in various calculations. "
                                            "Such as, tfidf, bm25, etc."),
                    "tf__method": "Method used in term frequency calculation",
                    "tf__double_norm_k": "Smooth parameter used by double norm term frequency method.",
                    "idf__method": "Method used in Inverse Document Frequency calculation",
                    "bert__avg_document_length": "Average number of tokens in a bert tokenized document.",
                    "bert__avg_sentence_length": "Average number of tokens in a bert tokenized sentences.",
                    "bert__drop_suffix": ("Whether to drop BERT generated suffixes in various calculations. "
                                          "Such as, tfidf, bm25, etc."),
                    "simple__avg_document_length": "Average token count in a simple tokenizer tokenized document.",
                    "simple__avg_sentence_length": "Average token count in a simple tokenizer tokenized sentences.",
                    "bm25__k1": "BM25 k1 parameter as defined in https://en.wikipedia.org/wiki/Okapi_BM25",
                    "bm25__b": "BM25 b parameter as defined in https://en.wikipedia.org/wiki/Okapi_BM25",
                    "bm25__delta": "BM25+ delta parameter as defined in https://en.wikipedia.org/wiki/Okapi_BM25",
                    }

    default_config = ConfigParser()
    default_config.read([Path(dirname(__file__)) / 'default.ini'])

    console = Console()

    table = Table(show_header=True, header_style="bold #2070b2")

    table.add_column("section")
    table.add_column("parameter_name")
    table.add_column("current_value")
    table.add_column("default_value")
    table.add_column("description", width=40)

    for sec in config.sections():
        if sec == section or section is None:
            for k in config[sec]:
                if config[sec][k] != default_config[sec][k]:
                    table.add_row(sec, k, f"[orange1]{config[sec][k]}[/orange1]", default_config[sec][k],
                                  descriptions.get(f"{sec}__{k}", ""))
                else:
                    table.add_row(sec, k, f"{config[sec][k]}", default_config[sec][k],
                                  descriptions.get(f"{sec}__{k}", ""))

    console.print(table)



from typing import List
from math import ceil

from sklearn.cluster import KMeans  # type: ignore
from sklearn.decomposition import PCA  # type: ignore
from sklearn.pipeline import Pipeline  # type: ignore

from ._base import ExtractiveSummarizer
from ..bblock import Sentences
from ..config import tokenizer_context


class KMeansSummarizer(ExtractiveSummarizer):
    tags = ExtractiveSummarizer.tags + ['cluster', 'ml']

    def __init__(self, n_clusters=2, random_state=42, normalize=True):
        super().__init__(normalize)
        self.n_clusters = n_clusters
        self.random_state = random_state

    def _predict(self, sentences: List[Sentences]):
        with tokenizer_context('bert', warning=True) as Doc:
            if len(sentences) == 0:
                raise ValueError(f"Ensure that document contains a few sentences for summarization")

            doc = Doc(sentences[0].document.raw)

            effective_n_clusters = min(self.n_clusters, len(doc))

            return 1 / (KMeans(n_clusters=effective_n_clusters, random_state=self.random_state).fit_transform(
                doc.bert_embeddings).min(axis=1) + 1e-10)


class AutoKMeansSummarizer(ExtractiveSummarizer):
    """Kmeans cluster automatically deciding on the number of clusters to be used based on document length."""

    tags = ExtractiveSummarizer.tags + ['cluster', 'ml']

    def __init__(self, n_cluster_to_length=0.05, min_n_cluster=2, random_state=42, normalize=True):
        super().__init__(normalize)

        self.n_cluster_to_length = n_cluster_to_length
        self.min_n_cluster = min_n_cluster
        self.random_state = random_state

    def _predict(self, sentences: List[Sentences]):
        with tokenizer_context('bert', warning=True) as Doc:
            if len(sentences) == 0:
                raise ValueError(f"Ensure that document contains a few sentences for summarization")

            doc = Doc(sentences[0].document.raw)

            effective_n_clusters = min(max(ceil(len(doc) * self.n_cluster_to_length), self.min_n_cluster), len(doc))

            return 1 / (KMeans(n_clusters=effective_n_clusters, random_state=self.random_state).fit_transform(
                doc.bert_embeddings).min(axis=1) + 1e-10)


class DecomposedKMeansSummarizer(ExtractiveSummarizer):
    """BERT embeddings are high in dimension and potentially carry redundant information that can cause
        overfitting or curse of dimensionality effecting in clustering embeddings.

        DecomposedKMeansSummarizer adds a PCA step (or any othe lsinear/non-linear dimensionality reduction technique)
         before clustering to obtain highest variance in vector fed into clustering
    """

    tags = ExtractiveSummarizer.tags + ['cluster', 'ml']

    def __init__(self, n_clusters=2, n_components=48, random_state=42, normalize=True):
        super().__init__(normalize)
        self.n_clusters = n_clusters
        self.n_components = n_components
        self.random_state = random_state

    def _predict(self, sentences: List[Sentences]):
        with tokenizer_context('bert', warning=True) as Doc:
            if len(sentences) == 0:
                raise ValueError(f"Ensure that document contains a few sentences for summarization")

            doc = Doc(sentences[0].document.raw)

            effective_n_clusters = min(self.n_clusters, len(doc))
            effective_n_components = min(self.n_components, len(doc))

            pipeline = Pipeline(
                [('pca', PCA(effective_n_components)),
                 ('kmeans', KMeans(effective_n_clusters, random_state=self.random_state))])

            return 1 / (pipeline.fit_transform(doc.bert_embeddings).min(axis=1) + 1e-10)



import gzip
import json
import sys
from os.path import expanduser, getsize
from pathlib import Path
from typing import Iterator

from rich.console import Console

console = Console()

tarballs = ["art_culture.jsonl.gz", "education.jsonl.gz", "horoscope.jsonl.gz", "life_food.jsonl.gz",
            "politics.jsonl.gz", "technology.jsonl.gz", "economics.jsonl.gz", "health.jsonl.gz",
            "life.jsonl.gz", "magazine.jsonl.gz", "sports.jsonl.gz", "travel.jsonl.gz"]

CATEGORIES = [tb.replace('.jsonl.gz', '') for tb in tarballs]

CORPUS_SIZE = 302936

__general_download_message__ = """Ensure that you have properly downloaded extended or tokenization corpus using

            python -m sadedegel.dataset.extended download --access-key xxx --secret-key xxxx
            python -m sadedegel.dataset.tokenization download --access-key xxx --secret-key xxxx

        Unfortunately due to data licensing issues we could not share data publicly. 
        Get in touch with sadedegel team to obtain a download key.

        """

__tokenization_download_message__ = """Ensure that you have properly downloaded tokenization corpus using

            python -m sadedegel.dataset.tokenization download --access-key xxx --secret-key xxxx

        Unfortunately due to data licensing issues we could not share data publicly. 
        Get in touch with sadedegel team to obtain a download key.

        """

__tokenization_version_message__ = "Ensure your dataset is in versioned format."

__tscorpus_validate__ = """It seems that your copy of TSCorpus is [red]corrupted[/red]. For validation run
 
            python -m sadedegel.dataset.tscorpus validate
            
        To download corpus again run
        
            python -m sadedegel.dataset.tscorpus download --access-key xxx --secret-key xxxx
"""


def fail_with_corpus_validation():
    console.log(__tscorpus_validate__)
    sys.exit(1)


def check_directory_structure(path: str) -> bool:
    base_dir = Path(expanduser(path))
    if not base_dir.exists():
        console.log(f"Dataset base directory ([bold red]{base_dir}[/bold red]) does not exist")

    elif not (base_dir / "tscorpus" / "raw").exists() or not (base_dir / "tscorpus" / "tokenized").exists():
        console.log(f"TsCorpus directory ([bold red]{base_dir}/tscorpus[/bold red]) does not exist")

    else:
        return True

    console.log(__general_download_message__)

    return False


def load_tscorpus_raw(data_home="~/.sadedegel_data") -> Iterator[str]:
    if not check_directory_structure(data_home):
        raise Exception("Ts Corpus validation error")

    for tarball in tarballs:
        with gzip.open(Path(expanduser(data_home)) / 'tscorpus' / 'raw' / tarball, 'rt') as fp:
            for line in fp:
                yield json.loads(line)


def load_tokenization_raw(data_home="~/.sadedegel_data") -> Iterator[str]:
    for d in load_tscorpus_raw(data_home):
        yield dict(id=d['id'], text=d['text'])


def load_classification_raw(data_home="~/.sadedegel_data") -> Iterator[str]:
    for d in load_tscorpus_raw(data_home):
        yield dict(id=d['id'], category=CATEGORIES.index(d['category']), text=d['text'])


def load_tokenization_tokenized(data_home="~/.sadedegel_data") -> Iterator[str]:
    if check_directory_structure(data_home):
        for tarball in tarballs:
            with gzip.open(Path(expanduser(data_home)) / 'tscorpus' / 'tokenized' / tarball, 'rt') as fp:
                for line in fp:
                    d = json.loads(line)
                    yield dict(id=d['id'], tokens=d['text'])
    else:
        return None


def raw_stats(data_home: str) -> int:
    sz = 0
    for f in (Path(expanduser(data_home)) / "tscorpus" / "raw").glob("*.gz"):
        sz += getsize(f)
    return sz


def tokenized_stats(data_home: str) -> int:
    sz = 0
    for f in (Path(expanduser(data_home)) / "tscorpus" / "tokenized").glob("*.gz"):
        sz += getsize(f)
    return sz


def check_and_display(data_home="~/.sadedegel_data"):
    if check_directory_structure(data_home):
        return dict(byte=dict(raw=raw_stats(data_home),
                              tokens=tokenized_stats(data_home)))



from math import ceil
from pathlib import Path
from os.path import dirname

from sklearn.linear_model import SGDClassifier
from sklearn.utils import shuffle
from sklearn.metrics import accuracy_score

from rich.console import Console

from joblib import dump, load as jl_load

import numpy as np

from ..dataset.tscorpus import load_classification_raw, CATEGORIES, CORPUS_SIZE
from ..extension.sklearn import TfidfVectorizer, OnlinePipeline, Text2Doc

console = Console()


def empty_model():
    return OnlinePipeline(
        [('text2doc', Text2Doc("icu")),
         ('tfidf', TfidfVectorizer(tf_method='log_norm', idf_method='smooth', drop_punct=True, drop_stopwords=False,
                                   lowercase=True)),
         ('sgd',
          SGDClassifier(penalty="l2", alpha=0.005190632263776186, loss="log", average=False, fit_intercept=True))])


def build(max_rows=-1, batch_size=10000):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    raw = load_classification_raw()
    df = pd.DataFrame.from_records(raw)
    df = shuffle(df, random_state=42)

    if max_rows > 0:
        df = df.sample(max_rows)

    np.random.seed(42)
    msk = np.random.rand(len(df)) < 0.9

    df_train = df[msk]  # random state is a seed value
    df_test = df[~msk]

    BATCH_SIZE = batch_size

    n_split = ceil(len(df_train) / BATCH_SIZE)
    console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

    batches = np.array_split(df_train, n_split)

    pipeline = empty_model()

    for batch in batches:
        pipeline.partial_fit(batch.text, batch.category, classes=[i for i in range(len(CATEGORIES))])

        y_pred = pipeline.predict(df_test.text)
        console.log(f"Accuracy on test: {accuracy_score(df_test.category, y_pred)}")

    console.log("Model build [green]DONE[/green]")

    model_dir = Path(dirname(__file__)) / 'model'

    model_dir.mkdir(parents=True, exist_ok=True)

    pipeline.steps[0][1].Doc = None

    dump(pipeline, (model_dir / 'news_classification.joblib').absolute(), compress=('gzip', 9))


def load(model_name="news_classification"):
    return jl_load(Path(dirname(__file__)) / 'model' / f"{model_name}.joblib")


if __name__ == '__main__':
    build(-1, batch_size=10_000)



import json
from math import ceil
from typing import List, Tuple

import numpy as np
from rich.live import Live
from rich.table import Table
from rich.console import Console
from sklearn.metrics import ndcg_score  # type: ignore

console = Console()


def topk(scores: dict, k: int = 3) -> List[Tuple[str, Tuple[float, float, float]]]:
    return sorted(scores.items(), key=lambda kv: kv[1][1], reverse=True)[:k]


def topk_table(scores: dict, i: int, total: int, k: int = 3):
    """Make a new table."""
    table = Table(title=f"{i}/{total} completed...")
    table.add_column("algorithm")
    table.add_column("parameters")
    table.add_column("ndcg (k=0.1)")
    table.add_column("ndcg (k=0.5)")
    table.add_column("ndcg (k=0.8)")

    for k, v in topk(scores, k):
        table.add_row(k[0], k[1], str(v[0]), str(v[1]), str(v[2]))

    return table


def grid_search(relevance, docs, summarize_class, parameter_space):
    scores = {}

    with Live(console=console, screen=True, auto_refresh=False) as live:
        for i, param in enumerate(parameter_space, 1):
            summarizer = summarize_class(**param)

            score_10, score_50, score_80 = [], [], []

            for y_true, d in zip(relevance, docs):
                y_pred = [summarizer.predict(d)]

                score_10.append(ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.1)))
                score_50.append(ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.5)))
                score_80.append(ndcg_score(y_true, y_pred, k=ceil(len(d) * 0.8)))

            scores[(summarizer.__class__.__name__, json.dumps(param))] = np.array(score_10).mean(), np.array(
                score_50).mean(), np.array(score_80).mean()

            live.update(topk_table(scores, i, len(parameter_space)), refresh=True)

    return scores



from setuptools import setup
from os.path import dirname
import sys


def version():
    sys.path.insert(0, dirname(__file__))

    from sadedegel.about import __version__

    return __version__


with open('prod.requirements.txt') as fp:
    install_requires = fp.read().splitlines()

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='sadedegel',
    # Needed to actually package something
    packages=['sadedegel', 'sadedegel.bblock',
              'sadedegel.dataset', 'sadedegel.dataset.extended', 'sadedegel.dataset.profanity',
              'sadedegel.dataset.tweet_sentiment', 'sadedegel.dataset.tscorpus',
              'sadedegel.summarize',
              'sadedegel.summarize.evaluate',
              'sadedegel.summarize.util', 'sadedegel.extension',
              'sadedegel.bblock.cli',
              'sadedegel.tokenize', 'sadedegel.ml',
              'sadedegel.server', 'sadedegel.metrics', 'sadedegel.prebuilt'],
    package_data={
        'sadedegel.dataset': ['raw/*.txt', 'sents/*.json', 'annotated/*.json'],
        'sadedegel.ml': ['model/sbd.pickle'],
        'sadedegel.prebuilt': ['model/*.joblib'],
        'sadedegel.bblock': ['data/bert/vocabulary.hdf5', 'data/simple/vocabulary.hdf5', 'data/icu/vocabulary.hdf5',
                             'data/stop-words.txt'],
        'sadedegel': ['default.ini']
    },
    # Needed for dependencies
    install_requires=install_requires,
    # *strongly* suggested for sharings
    version=version(),
    python_requires='>=3.6',
    entry_points='''
        [console_scripts]
        sadedegel=sadedegel.__main__:cli
        sadedegel-dataset=sadedegel.dataset.__main__:cli
        sadedegel-dataset-extended=sadedegel.dataset.extended.__main__:cli
        sadedegel-dataset-tscorpus=sadedegel.dataset.tscorpus.__main__:cli
        sadedegel-bblock=sadedegel.bblock.cli.__main__:cli
        sadedegel-summarize=sadedegel.summarize.evaluate.__main__:cli
        sadedegel-sbd=sadedegel.tokenize.__main__:cli
        sadedegel-server=sadedegel.server.__main__:server
        sadedegel-build-vocabulary=sadedegel.bblock.cli.__main__:build_vocabulary
    ''',
    extras_require={
        'w2v': ['gensim'],
        'bert': ['torch==1.5.1', 'transformers==3.0.0']
    }
)



from ._sent import RegexpSentenceTokenizer, NLTKPunctTokenizer  # noqa: F401
from ..bblock.util import tr_lower, tr_upper, __tr_upper__, __tr_lower__  # noqa: F401

# This is kepts for backward compatibility
# from .. import Doc, Sentences  # noqa: F401

from .. import Doc, Sentences



from typing import List
import numpy as np

from ..tokenize import Sentences
from ._base import ExtractiveSummarizer

from ..config import load_config


class TFIDFSummarizer(ExtractiveSummarizer):
    tags = ExtractiveSummarizer.tags + ['ml', 'tfidf', 'ir']

    def __init__(self, tf_method=None, idf_method=None, drop_stopwords=False, drop_suffix=False, drop_punct=False,
                 lowercase=False, normalize=True, kwargs={}):
        super().__init__(normalize)

        cfg = load_config()

        self.drop_stopwords = drop_stopwords
        self.drop_suffix = drop_suffix
        self.drop_punct = drop_punct
        self.lowercase = lowercase

        self.tf_method = tf_method if tf_method is not None else cfg['tf']['method']
        self.idf_method = idf_method if idf_method is not None else cfg['idf']['method']

        self.kwargs = kwargs

    def _predict(self, sents: List[Sentences]):
        return np.array([sent.get_tfidf(tf_method=self.tf_method,
                                        idf_method=self.idf_method,
                                        drop_stopwords=self.drop_stopwords,
                                        drop_suffix=self.drop_suffix,
                                        drop_punct=self.drop_punct, **self.kwargs).sum() for sent in sents])



import sys
from collections import Counter
from collections import defaultdict
from itertools import islice
from itertools import tee
from multiprocessing import cpu_count
from typing import Tuple

import click
import numpy as np
from rich.console import Console
from rich.progress import track
from rich.table import Table
from tqdm import tqdm

from ..vocabulary import VocabularyCounter
from ..word_tokenizer_helper import word_tokenize, ICUTokenizerHelper
from ...config import tokenizer_context
from ...dataset.extended import load_extended_sents_corpus
from ...dataset.tscorpus import load_tokenization_tokenized, load_tokenization_raw, fail_with_corpus_validation, \
    CORPUS_SIZE


def get_w2v_model(size: int = 100, min_count=3, skip_gram=True, workers: int = cpu_count(), seed=42):
    try:
        from gensim.models import Word2Vec
    except ImportError:
        print(
            ("Error in importing gensim module. "
             "Ensure that you run 'pip install sadedegel[gensim]' to enable word2vec training"),
            file=sys.stderr)
        sys.exit(1)

    return Word2Vec(size=size, workers=workers, sg=skip_gram, min_count=min_count, iter=iter, seed=seed)


console = Console()


def tok_eval(tokenizer, limit=None) -> Tuple[str, float, float]:
    """Evaluate tokenizers on a downsized version of tokenization dataset.

    :type tokenizer: str
    :param limit: Top-K documents for evaluation
    :type limit: int

    :return: IoU score between list of true tokens and list of tokenized tokens.
    """
    if tokenizer == "bert":
        try:
            import torch
            from transformers import AutoTokenizer
        except ImportError:
            console.print(
                ("Error in importing transformers module. "
                 "Ensure that you run 'pip install sadedegel[bert]' to use BERT features."),
                file=sys.stderr)
            sys.exit(1)

        toker = AutoTokenizer.from_pretrained("dbmdz/bert-base-turkish-cased").tokenize

    elif tokenizer == "simple":
        toker = word_tokenize
    else:
        toker = ICUTokenizerHelper()

    console.log(f"[bold blue]Evaluating {tokenizer} tokenizer...")

    if limit:
        raw = islice(load_tokenization_raw(), limit)
        tok = islice(load_tokenization_tokenized(), limit)
    else:
        raw = load_tokenization_raw()
        tok = load_tokenization_tokenized()

    macro = []
    micro = defaultdict(int)
    wmacro = []
    wmicro = defaultdict(int)
    for i, (raw_doc, tokenized_doc) in tqdm(enumerate(zip(raw, tok)), desc="Evaluating...",
                                            total=limit if limit else CORPUS_SIZE):
        if raw_doc['id'] != tokenized_doc['id']:
            fail_with_corpus_validation()

        true_tokens = Counter(tokenized_doc['tokens'])

        tokens = toker(raw_doc['text'])

        if tokenizer == 'bert':
            fixed_hashtags = []
            for tok in tokens:
                if "##" not in tok:
                    fixed_hashtags.append(tok)
                else:
                    merged = fixed_hashtags[-1] + tok.split("##")[1]
                    fixed_hashtags.pop(-1)
                    fixed_hashtags.append(merged)
            tokens = fixed_hashtags

        pred_tokens = Counter(tokens)

        union = sum((pred_tokens | true_tokens).values())
        intersection = sum((pred_tokens & true_tokens).values())

        wmacro.append(intersection / union)

        wmicro['intersection'] += intersection
        wmicro['union'] += union

        union = len(set(pred_tokens) | set(true_tokens))
        intersection = len(set(pred_tokens) & set(true_tokens))

        macro.append(intersection / union)

        micro['intersection'] += intersection
        micro['union'] += union

    # console.log(f"{np.mean(macro):.4f}", f"{micro['intersection'] / micro['union']:.4f}")
    # console.log(f"Weighted: {np.mean(wmacro):.4f}", f"{wmicro['intersection'] / wmicro['union']:.4f}")

    return tokenizer, np.mean(macro), micro['intersection'] / micro['union']


@click.group()
def cli():
    pass


@click.option("--tokenizers", "-t", help="List of tokenizers to evaluate", multiple=True)
@click.option("--limit", "-s", help="Evaluation set size", type=int)
@cli.command()
def tokenizer_evaluate(tokenizers, limit):
    """Tokenizer evaluation"""

    result = []
    for tok in tokenizers:
        result.append(tok_eval(tok, limit))

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Tokenizer")
    table.add_column("IoU (macro)")
    table.add_column("IoU (micro)")

    for row in result:
        table.add_row(row[0], f"{row[1]:.4f}", f"{row[2]:.4f}")

    console.print(table)


def sentence_iter(word_tokenizer, max_doc=None):
    if max_doc is None:
        corpus = load_extended_sents_corpus()
    else:
        corpus = islice(load_extended_sents_corpus(), max_doc)

    with tokenizer_context(word_tokenizer) as Doc:
        for d in corpus:
            doc = Doc.from_sentences(d['sentences'])

            for sent in doc:
                yield [t.lower_ for t in sent if not t.is_punct]


@cli.command()
@click.option('--max-doc', help="Maximum number of documents in extended corpus", type=int, default=None)
@click.option('--min-df', help="Minimum document frequency of a word to be included in", default=3)
@click.option('--min-freq', help="Minimum term frequency", default=3)
@click.option('--word-tokenizer', "-t", type=click.Choice(['bert', "icu", "simple"], case_sensitive=False),
              help="Word tokenizer to be used in building vocabulary.", default='bert')
@click.option('--w2v/--no-w2v', default=True, help="Train word embeddings")
@click.option('--w2v-num-epoch', type=int, help="Number of epochs in word2vec training", default=10)
@click.option('--w2v-skip-gram', type=int, help='Skip Gram or CBOW. Defaults to True for Skip Gram', default=True)
@click.option('--w2v-size', type=int, help='Dimension of word vectors', default=100)
def build_vocabulary(max_doc, min_df, min_freq, word_tokenizer, w2v, w2v_num_epoch, w2v_skip_gram, w2v_size):
    """Build vocabulary & optionally word embeddings"""

    total = sum(1 for _ in sentence_iter(word_tokenizer, max_doc))

    counter_cs = VocabularyCounter(word_tokenizer, case_sensitive=True, min_tf=min_freq, min_df=min_df)
    counter = VocabularyCounter(word_tokenizer, case_sensitive=False, min_tf=min_freq, min_df=min_df)

    if w2v:
        model = get_w2v_model(w2v_size, min_freq, w2v_skip_gram)
        console.log("Building vocabulary...")
        model.build_vocab(sentence_iter(word_tokenizer, max_doc))

        for _ in track(range(w2v_num_epoch), total=w2v_num_epoch, description="Building word vectors..."):
            model.train(sentences=sentence_iter(word_tokenizer, max_doc),
                        epochs=1,
                        total_examples=total,
                        report_delay=1)

    click.secho(click.style(f"...Frequency calculation over extended dataset", fg="blue"))
    with tokenizer_context(word_tokenizer) as Doc:
        if max_doc is None:
            corpus = load_extended_sents_corpus()
        else:
            corpus = islice(load_extended_sents_corpus(), max_doc)

        corpus, X = tee(corpus, 2)

        for i, d in track(enumerate(corpus),
                          total=sum(1 for _ in X),
                          description="Building vocabulary..."):
            doc = Doc.from_sentences(d['sentences'])

            for sent in doc:
                for word in sent.tokens:
                    counter.inc(word, i, 1)
                    counter_cs.inc(word, i, 1)

    counter = counter.prune()
    counter_cs = counter_cs.prune()

    if w2v:
        counter.to_hdf5(model)
    else:
        counter.to_hdf5()
    counter_cs.to_hdf5()


if __name__ == '__main__':
    cli()



import warnings
from collections import defaultdict
from os.path import dirname
from pathlib import Path

import h5py
import numpy as np
from cached_property import cached_property
from rich.console import Console

from .util import tr_lower, normalize_tokenizer_name

console = Console()


class InvalidTokenizer(Exception):
    """Invalid tokenizer name"""


def vocabulary_file(tokenizer: str, verify_exists=True):
    normalized_name = normalize_tokenizer_name(tokenizer)

    if normalized_name not in ['bert', 'icu', 'simple']:
        raise InvalidTokenizer(
            (f"Currently only valid tokenizers are BERT, ICU Tokenizer for vocabulary generation."
             " {normalized_name} found"))

    vocab_file = Path(dirname(__file__)) / 'data' / normalized_name / 'vocabulary.hdf5'

    if not vocab_file.exists() and verify_exists:
        raise FileNotFoundError(f"Vocabulary file for {tokenizer} ({normalized_name}) tokenizer not found.")

    return vocab_file


class VocabularyCounter:
    def __init__(self, tokenizer, case_sensitive=True, min_tf=1, min_df=1):
        self.tokenizer = tokenizer

        self.doc_counter = defaultdict(set)
        self.doc_set = set()

        self.term_freq = defaultdict(int)

        self.min_tf = min_tf
        self.min_df = min_df
        self.case_sensitive = case_sensitive

    def inc(self, word: str, document_id: int, count: int = 1):
        if self.case_sensitive:
            w = word
        else:
            w = tr_lower(word)

        self.doc_counter[w].add(document_id)
        self.doc_set.add(document_id)
        self.term_freq[w] += count

    def add_word_to_doc(self, word: str, document_id: int):
        """Implemented for backward compatibility"""

        self.inc(word, document_id, 1)

    @property
    def vocabulary_size(self):
        return len(self.term_freq)

    @property
    def document_count(self):
        return len(self.doc_set)

    def prune(self):

        to_remove = []

        for w in self.term_freq:
            if self.term_freq[w] < self.min_tf or len(self.doc_counter[w]) < self.min_df:
                to_remove.append(w)

        for w in to_remove:
            del self.doc_counter[w]
            del self.term_freq[w]

        console.log(
            f"{len(to_remove)} terms (case sensitive={self.case_sensitive}) are pruned by tf (>= {self.min_tf}) or df filter(>= {self.min_df})")

        return self

    def df(self, w: str):
        if self.case_sensitive:
            return len(self.doc_counter[w])
        else:
            return len(self.doc_counter[tr_lower(w)])

    def tf(self, w: str):
        if self.case_sensitive:
            return self.term_freq[w]
        else:
            return self.term_freq[tr_lower(w)]

    def to_hdf5(self, w2v=None):
        with h5py.File(vocabulary_file(self.tokenizer, verify_exists=False), "a") as fp:
            if self.case_sensitive:
                group = fp.create_group("form_")
            else:
                group = fp.create_group("lower_")

            words = sorted(list(self.term_freq.keys()), key=lambda w: tr_lower(w))

            group.attrs['size'] = len(words)
            group.attrs['document_count'] = len(self.doc_set)
            group.attrs['tokenizer'] = self.tokenizer
            group.attrs['min_tf'] = self.min_tf
            group.attrs['min_df'] = self.min_df

            if w2v is not None:
                group.attrs['vector_size'] = w2v.vector_size

                group.create_dataset("vector", data=np.array(
                    [w2v[w] if w in w2v else np.zeros(w2v.vector_size) for w in words]).astype(
                    np.float32),
                                     compression="gzip",
                                     compression_opts=9)
                group.create_dataset("has_vector", data=np.array([w in w2v in w2v for w in words]),
                                     compression="gzip",
                                     compression_opts=9)

            group.create_dataset("word", data=words, compression="gzip", compression_opts=9)
            group.create_dataset("df", data=np.array([self.df(w) for w in words]), compression="gzip",
                                 compression_opts=9)
            group.create_dataset("tf", data=np.array([self.tf(w) for w in words]), compression="gzip",
                                 compression_opts=9)

        console.print(f"|D|: {self.document_count}, |V|: {self.vocabulary_size} (case sensitive={self.case_sensitive})")


class Vocabulary:

    def __init__(self, tokenizer):
        self.tokenizer = tokenizer

        self.file_name = vocabulary_file(tokenizer)
        self._df = None
        self._df_cs = None
        self._has_vector = None
        self._vector = None

        self.dword_cs = None
        self.dword = None

    @cached_property
    def size_cs(self) -> int:
        with h5py.File(self.file_name, "r") as fp:
            return fp['form_'].attrs['size']

    @cached_property
    def size(self) -> int:
        with h5py.File(self.file_name, "r") as fp:
            return fp['lower_'].attrs['size']

    def __len__(self):
        return self.size

    def id_cs(self, word: str, default: int = -1):
        if self.dword_cs is None:
            with h5py.File(self.file_name, "r") as fp:
                self.dword = dict((b.decode("utf-8"), i) for i, b in enumerate(list(fp['lower_']['word'])))
                self.dword_cs = dict((b.decode("utf-8"), i) for i, b in enumerate(list(fp['form_']['word'])))

        return self.dword_cs.get(word, default)

    def id(self, word: str, default: int = -1):
        if self.dword is None:
            with h5py.File(self.file_name, "r") as fp:
                self.dword = dict((b.decode("utf-8"), i) for i, b in enumerate(list(fp['lower_']['word'])))
                self.dword_cs = dict((b.decode("utf-8"), i) for i, b in enumerate(list(fp['form_']['word'])))

        return self.dword.get(tr_lower(word), default)

    def df(self, word: str):

        i = self.id(word)

        if i == -1:
            return 0
        else:
            if self._df is None:
                with h5py.File(self.file_name, "r") as fp:
                    self._df = np.array(fp['lower_']['df'])

            return self._df[i]

    def df_cs(self, word: str):

        i = self.id_cs(word)

        if i == -1:
            return 0
        else:
            if self._df_cs is None:
                with h5py.File(self.file_name, "r") as fp:
                    self._df_cs = np.array(fp['form_']['df'])

            return self._df_cs[i]

    def has_vector(self, word: str):
        with h5py.File(self.file_name, "r") as fp:
            if "has_vector" in fp['lower_']:
                i = self.id(word)

                if i == -1:
                    return False
                else:
                    if self._has_vector is None:
                        self._has_vector = np.array(fp['lower_']['has_vector'])

                    return self._has_vector[i]
            else:
                return False

    def vector(self, word: str):
        # TODO: Performance improvement required
        with h5py.File(self.file_name, "r") as fp:
            if "vector" in fp['lower_']:
                i = self.id(word)

                if i == -1:
                    return False
                else:
                    if self._vector is None:
                        self._vector = np.array(fp['lower_']['vector'])

                    return self._vector[i, :]
            else:
                return False

    @cached_property
    def document_count(self):
        with h5py.File(self.file_name, "r") as fp:
            return fp['form_'].attrs['document_count']



import sys
import warnings
from abc import ABC, abstractmethod
from typing import List

from cached_property import cached_property
from rich.console import Console

from .util import normalize_tokenizer_name
from .vocabulary import Vocabulary
from .word_tokenizer_helper import word_tokenize, ICUTokenizerHelper
from ..about import __version__

console = Console()


class WordTokenizer(ABC):
    __instances = {}

    def __init__(self):
        self._vocabulary = None

    @abstractmethod
    def _tokenize(self, text: str) -> List[str]:
        pass

    @abstractmethod
    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        pass

    def __call__(self, sentence: str) -> List[str]:
        return self._tokenize(str(sentence))

    @staticmethod
    def factory(tokenizer_name: str):
        normalized_name = normalize_tokenizer_name(tokenizer_name)
        if normalized_name not in WordTokenizer.__instances:
            if normalized_name == "bert":
                WordTokenizer.__instances[normalized_name] = BertTokenizer()
            elif normalized_name == "simple":
                warnings.warn(
                    ("Note that SimpleTokenizer is pretty new in sadedeGel. "
                     "If you experience any problems, open up a issue "
                     "(https://github.com/GlobalMaksimum/sadedegel/issues/new)"))
                WordTokenizer.__instances[normalized_name] = SimpleTokenizer()
            elif normalized_name == "icu":
                WordTokenizer.__instances[normalized_name] = ICUTokenizer()
            else:
                raise Exception(
                    (f"No word tokenizer type match with name {tokenizer_name}."
                     " Use one of 'bert-tokenizer', 'SimpleTokenizer', etc."))

        return WordTokenizer.__instances[normalized_name]


class BertTokenizer(WordTokenizer):
    __name__ = "BertTokenizer"

    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        return self.tokenizer.convert_tokens_to_ids(tokens)

    def __init__(self):
        super(BertTokenizer, self).__init__()

        self.tokenizer = None

    def _tokenize(self, text: str) -> List[str]:
        if self.tokenizer is None:
            try:
                import torch
                from transformers import AutoTokenizer
            except ImportError:
                console.print(
                    ("Error in importing transformers module. "
                     "Ensure that you run 'pip install sadedegel[bert]' to use BERT features."))
                sys.exit(1)
            self.tokenizer = AutoTokenizer.from_pretrained("dbmdz/bert-base-turkish-cased")

        return self.tokenizer.tokenize(text)

    @cached_property
    def vocabulary(self):
        try:
            return Vocabulary("bert")
        except FileNotFoundError:
            console.print("[red]bert[/red] vocabulary file not found.")

            return None


class SimpleTokenizer(WordTokenizer):
    __name__ = "SimpleTokenizer"

    def __init__(self):
        super(SimpleTokenizer, self).__init__()
        self.tokenizer = word_tokenize

    def _tokenize(self, text: str) -> List[str]:
        return self.tokenizer(text)

    def convert_tokens_to_ids(self, ids: List[str]) -> List[int]:
        raise NotImplementedError("convert_tokens_to_ids is not implemented for SimpleTokenizer yet. Use BERTTokenizer")

    @cached_property
    def vocabulary(self):
        try:
            return Vocabulary("simple")
        except FileNotFoundError:
            console.print("[red]simple[/red] vocabulary file not found.")

            return None


class ICUTokenizer(WordTokenizer):
    __name__ = "ICUTokenizer"

    def __init__(self):
        super(ICUTokenizer, self).__init__()
        self.tokenizer = ICUTokenizerHelper()

    def _tokenize(self, text: str) -> List[str]:
        return self.tokenizer(text)

    def convert_tokens_to_ids(self, ids: List[str]) -> List[int]:
        raise NotImplementedError("convert_tokens_to_ids is not implemented for SimpleTokenizer yet. Use BERTTokenizer")

    @cached_property
    def vocabulary(self):
        try:
            return Vocabulary("icu")
        except FileNotFoundError:
            console.print("[red]icu[/red] vocabulary file not found.")

            return None


def get_default_word_tokenizer() -> WordTokenizer:
    if tuple(map(int, __version__.split('.'))) < (0, 17):
        warnings.warn(
            ("get_default_word_tokenizer is deprecated and will be removed by 0.17. "
             "Use `sadedegel config` to get default configuration. "
             "Use ~/.sadedegel/user.ini to update default tokenizer."),
            DeprecationWarning,
            stacklevel=2)
    else:
        raise Exception("Remove get_default_word_tokenizer before release.")

    return WordTokenizer.factory(BertTokenizer.__name__)



import re
import string
from typing import List

from icu import Locale, BreakIterator

from .util import space_pad

metatoken_telno = "phone"
metatoken_cctelno = "shortphone"
metatoken_hhmm = "hhmm"
metatoken_plate = "plate"
metatoken_ddmmyyyy = "ddmmyyyy"

mail_re = re.compile(
    r"[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*")

website_re = re.compile(
    r'\b(http[s]?\:\/\/)?[a-zı0-9]{3,}([./][a-zı0-9]*)+[./a-zı0-9]+(:[a-zı0-9]+)?')
phone_re = re.compile(
    r'\b(\(?0[ -]?[0-9]{3}\)?[ -]?[0-9]{3}[ -]?[0-9]{2}[ -]?[0-9]{2}|0 ?[(]{1}[0-9]{3}[)][ -][0-9]{3}[ -][0-9]{2}[ -][0-9]{2}|[2|5][0-9]{2}[ -]?[0-9]{3}[ -]?[0-9]{2}[ -]?[0-9]{2})|((0\s*)?21[26]\s+\d{3}\s+\d{2}\s+\d{2})')
cc_phone_re = re.compile(r'([444]{3}[ -]?[0-9]{1,2}[ -]?[0-9]{2,3})|(444\s*\d{2}\s*\d{2})')
# time_re = re.compile(
#    r'([0-9]{1,2}[:., ][0-9]{2}[ ]{0,2}[pm|am]|[0-9]{1,2}[:., ][0-9]{2})|((0[0-9]|1[0-9]|2[0-3])(:|\s+|\.)[0-5][0-9](:[0-5][0-9])?)')

time_re = re.compile(
    r"\b(?P<hour>1[0-9]|2[0-3]|0?[0-9])[:\.,](?P<minute>[1-5][0-9]|0?[0-9])(?:[:\.,](?P<second>[1-5][0-9]|0?[0-9]))?(?:[:\.,\s+](?P<pm_am>[ap]\.?m\.?))?")
carplate_re = re.compile(r"(?P<city>(0[1-9])|([1-7][0-9])|(8[01]))\s*(?P<mid>[A-Z][A-Z0-9]{1,4})\s*(?P<suffix>\d{2,4})")
# date_re = re.compile(
#    r'([0-9]{1,2}[./\-: ][0-9]{1,2}[./\-: ][0-9]{2,4}|[0-9]{1,2}[ ][A-ZİŞĞÖÜa-zışğöü]{1,10}[ ][0-9]{2,4}|[A-ZİŞĞÖÜa-zışğöü]{1,10}[ ][0-9]{1,2}[,][ ][0-9]{2,4})|(20\d{2}-\d{2}-\d{2})|(\d{2}[\.-:,]\d{2}[\.-:,]20\d{2})')
date_1_re = re.compile(
    r'\b(?P<day>0?[0-9]|1[0-9]|2[0-9]|3[01])(?:\s+|[:\-\./])(?P<month>oca(?:k)?|[sş]u(?:bat)?|mar(?:t)?|n[ıi]s(?:an)?|may(?:[ıi]s)?|haz(?:[iı]ran)?|tem(?:muz)?|a[gğ]u(?:stos)?|eyl(?:[uü]l)?|ek[iı](?:m)?|kas(?:[iı]m)?|ara(?:l[iı]k)?)(?:\s+|[:\-\./])(?P<year>\d{4})?',
    flags=re.IGNORECASE)
date_2_re = re.compile(
    r'\b(?P<day>0?[0-9]|1[0-9]|2[0-9]|3[01])(?:\s+|[:\-\./])(?P<month>(?:1[0-2]|0?[1-9]))(?:\s+|[:\-\./])(?P<year>\d{4})',
    flags=re.IGNORECASE)
space = re.compile(r"\s+")
token_re = re.compile(r"[@]?[a-zA-ZşŞiİıIğĞüÜöÖçÇ0-9]+|[\.]")

puncts = string.punctuation + '”“’‘…'


class Compose:
    def __init__(self, *funcs):
        self.debug = False
        self.funcs = funcs[0]

    def __call__(self, arg):
        if self.debug:
            print(arg)
        res = self.funcs[0](arg)

        if self.debug:
            print(res)

        for i in range(1, len(self.funcs)):
            res = self.funcs[i](res)
            if self.debug:
                print(self.funcs[i], res)

        return res


def compose(*funcs):
    return Compose(funcs)


c = compose(
    lambda t: mail_re.sub("email", t),
    lambda t: website_re.sub("website", t),
    lambda t: phone_re.sub(space_pad(metatoken_telno), t),
    lambda t: cc_phone_re.sub(space_pad(metatoken_cctelno + " " + metatoken_telno), t),
    lambda t: date_1_re.sub(space_pad(metatoken_ddmmyyyy), t),
    lambda t: date_2_re.sub(space_pad(metatoken_ddmmyyyy), t),
    lambda t: time_re.sub(space_pad(metatoken_hhmm), t),
    lambda t: carplate_re.sub(space_pad(metatoken_plate), t),

    lambda t: space.sub(" ", t)
)

prefix_re = re.compile('^[-]+')


def tokenize_substring(text: str):
    for t in token_re.findall(text):
        yield prefix_re.sub('', t)


def word_tokenize_iter(text: str):
    for substring in c(text).split(' '):
        for t in tokenize_substring(substring):
            if len(t) > 0:
                yield t


def word_tokenize(text: str):
    return list(word_tokenize_iter(text))


class ICUTokenizerHelper:
    def __init__(self):
        self.locale = Locale("tr")
        self.breakor = BreakIterator.createWordInstance(self.locale)

    def __call__(self, text: str) -> List[str]:
        self.breakor.setText(text)

        parts = []
        p0 = 0
        for p1 in self.breakor:
            part = text[p0:p1].strip()
            if len(part) > 0:
                parts.append(part)
            p0 = p1

        return parts
